@@ -3,21 +3,27 @@ __author__ = "Akshay Mane"
 import requests
 import urllib3
 import logging
+
 # Disable insecure connections warnings
 urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
 
 
 class FortiManager:
-    def __init__(self, host, username="admin", password="admin", adom="root", verify=False):
-        protocol = "https"
+    """
+    This class will include all the methods used for executing the api calls on FortiManager.
+    """
+
+    def __init__(self, host, username="admin", password="admin", adom="root", protocol="https", verify=True):
+        self.protocol = protocol
         self.host = host
         self.username = username
         self.password = password
         self.adom = adom
-        self.sessionid = "null"
+        self.sessionid = None
+        self.session = None
         self.verify = verify
-        if not self.verify:
-            protocol = "http"
+        if protocol == "http":
+            self.verify = False
         self.base_url = f"{protocol}://{self.host}/jsonrpc"
 
     # Login Method
@@ -26,26 +32,26 @@ class FortiManager:
         Log in to FortiManager with the details provided during object creation of this class
         :return: Session
         """
-        session = requests.session()
-        payload = \
-            {
-                "method": "exec",
-                "params":
-                    [
-                        {
-                            "data": {
-                                "passwd": self.password,
-                                "user": self.username
-                            },
-                            "url": "sys/login/user"
-                        }
-                    ],
-                "session": self.sessionid
-            }
-        payload = repr(payload)
-        login = session.post(url=self.base_url, data=payload, verify=self.verify)
-        self.sessionid = login.json()['session']
-        return session
+        if self.sessionid is None:
+            self.session = requests.session()
+            payload = \
+                {
+                    "method": "exec",
+                    "params":
+                        [
+                            {
+                                "data": {
+                                    "passwd": self.password,
+                                    "user": self.username
+                                },
+                                "url": "sys/login/user"
+                            }
+                        ],
+                    "session": self.sessionid
+                }
+            login = self.session.post(url=self.base_url, json=payload, verify=self.verify)
+            self.sessionid = login.json()['session']
+        return self.session
 
     def logout(self):
         """
@@ -64,8 +70,7 @@ class FortiManager:
                     ],
                 "session": self.sessionid
             }
-        payload = repr(payload)
-        logout = session.post(url=self.base_url, data=payload, verify=self.verify)
+        logout = session.post(url=self.base_url, json=payload, verify=self.verify)
         return logout.json()["result"]
 
     # Adoms Methods
@@ -91,10 +96,65 @@ class FortiManager:
                     ],
                 "session": self.sessionid
             }
-        payload = repr(payload)
-        get_adoms = session.post(url=self.base_url, data=payload, verify=False)
+        get_adoms = session.post(url=self.base_url, json=payload, verify=False)
         return get_adoms.json()["result"]
 
+    def get_devices(self):
+        """
+        :return: returns list of devices added in FortiManager
+        """
+        session = self.login()
+        payload = {"method": "get", "params": [{"url": f"/dvmdb/adom/{self.adom}/device/"}]}
+        payload.update({"session": self.sessionid})
+        get_devices = session.post(url=self.base_url, json=payload, verify=False)
+        return get_devices.json()
+
+    def add_device(self, ip_address, username, password, name, description=False):
+        session = self.login()
+        payload = \
+            {
+                "method": "exec",
+                "params": [
+                    {"url": "dvm/cmd/add/device",
+                     "data": {"adom": f"{self.adom}", "flags": ["create_task", "nonblocking"],
+                              "device": {"adm_pass": f"{password}", "adm_usr": f"{username}", "desc": f"{description}",
+                                         "ip": f"{ip_address}",
+                                         "name": f"{name}", "mgmt_mode": 3}}}]}
+        payload.update({"session": self.sessionid})
+        add_device = session.post(url=self.base_url, json=payload, verify=self.verify)
+        return add_device.json()
+
+    def add_model_device(self, device_name, serial_no, username="admin", password=""):
+        session = self.login()
+        payload = {
+            "method": "exec",
+            "params": [
+                {
+                    "url": "dvm/cmd/add/device",
+                    "data": {
+                        "adom": self.adom,
+                        "flags": [
+                            "create_task",
+                            "nonblocking"
+                        ],
+                        "device": {
+                            "name": device_name,
+                            "adm_usr": username,
+                            "adm_pass": password,
+                            "flags": 67371040,
+                            "sn": serial_no,
+                            "platform_str": "FortiGate-VM64",
+                            "os_ver": 6,
+                            "mr": 2
+                        }
+                    }
+                }
+            ]
+        }
+        payload.update({"session": self.sessionid})
+        add_model_device = session.post(url=self.base_url, json=payload, verify=False)
+        return add_model_device.json()["result"]
+
     # Policy Package Methods
     def get_policy_packages(self, name=False):
         """
@@ -117,8 +177,7 @@ class FortiManager:
                     ],
                 "session": self.sessionid
             }
-        payload = repr(payload)
-        get_packages = session.post(url=self.base_url, data=payload, verify=False)
+        get_packages = session.post(url=self.base_url, json=payload, verify=False)
         return get_packages.json()["result"]
 
     def add_policy_package(self, name):
@@ -144,10 +203,72 @@ class FortiManager:
                     ],
                 "session": self.sessionid
             }
-        payload = repr(payload)
-        add_package = session.post(url=self.base_url, data=payload, verify=False)
+        add_package = session.post(url=self.base_url, json=payload, verify=False)
         return add_package.json()["result"]
 
+    def add_install_target(self, device_name, pkg_name, vdom="root"):
+        """
+        Add a device to installation target list of the policy package
+        :param device_name: name of the device
+        :param pkg_name: name of the policy package
+        :param vdom: name of the vdom (default=root)
+        :return: returns response from FortiManager api whether is was a success or failure.
+        """
+        session = self.login()
+        payload = \
+            {"method": "add",
+             "params": [{"url": f"pm/pkg/adom/{self.adom}/{pkg_name}/scope member",
+                         "data": [{"name": f"{device_name}",
+                                   "vdom": f"{vdom}"}]}]}
+        payload.update({"session": self.sessionid})
+        add_installation_target = session.post(url=self.base_url, json=payload, verify=self.verify)
+        return add_installation_target.json()
+
+    def get_meta_data(self):
+        """
+        Get all the meta tags present in the FortiManager
+        :return: returns meta tags present in FortiManager
+        """
+        session = self.login()
+        payload = {"method": "get", "params": [{"url": "/dvmdb/_meta_fields/device"}]}
+        payload.update({"session": self.sessionid})
+        get_meta = session.post(url=self.base_url, json=payload, verify=self.verify)
+        return get_meta.json()
+
+    def add_meta_data(self, name, importance=0, status=1):
+        """
+        Add a meta tag in the FortiManager.
+        :param name: name of the meta tag
+        :param importance: importance of meta tag
+        :param status: status of meta tag whether it should be active(1) or disabled(0)
+        :return: returns response from FortiManager API whether the request was successful or not.!
+        """
+        session = self.login()
+        payload = {"method": "add",
+                   "params": [
+                       {"url": "/dvmdb/_meta_fields/device",
+                        "data": {"importance": importance, "length": 255, "name": f"{name}",
+                                 "status": status}}]}
+        payload.update({"session": self.sessionid})
+        get_meta = session.post(url=self.base_url, json=payload, verify=self.verify)
+        return get_meta.json()
+
+    def assign_meta_to_device(self, device, meta_name, meta_value):
+        """
+        Assign a meta tag to the device
+        :param device: name of the device
+        :param meta_name: name of the meta tag
+        :param meta_value: value of the meta tag
+        :return: returns response from FortiManager API whether the request was successful or not.!
+        """
+        session = self.login()
+        payload = {"method": "update",
+                   "params": [{"url": f"/dvmdb/adom/root/device/{device}",
+                               "data": {"name": f"{device}", "meta fields": {f"{meta_name}": f"{meta_value}"}}}]}
+        payload.update({"session": self.sessionid})
+        assign_meta = session.post(self.base_url, json=payload, verify=self.verify)
+        return assign_meta.json()
+
     # Firewall Object Methods
     def get_firewall_address_objects(self, name=False):
         """
@@ -168,11 +289,10 @@ class FortiManager:
                 ],
                 "session": self.sessionid
             }
-        payload = repr(payload)
-        get_address_objects = session.post(url=self.base_url, data=payload, verify=self.verify)
+        get_address_objects = session.post(url=self.base_url, json=payload, verify=self.verify)
         return get_address_objects.json()["result"]
 
-    def add_firewall_address_object(self, name, associated_interface="any", subnet=list, object_type=0,
+    def add_firewall_address_object(self, name, subnet: list, associated_interface="any", object_type=0,
                                     allow_routing=0):
         """
         Create an address object using provided info
@@ -186,28 +306,68 @@ class FortiManager:
         session = self.login()
         payload = {
             "method": "add",
-            "params": [
-                {
-                    "data": {
-                        "allow-routing": allow_routing,
-                        "associated-interface": associated_interface,
-                        "name": name,
-                        "subnet": subnet,
-                        "type": object_type
-                    },
-                    "url": f"pm/config/adom/{self.adom}/obj/firewall/address"
-                }
-            ],
-            "session": self.sessionid
-        }
-        payload = repr(payload)
-        add_address_objects = session.post(url=self.base_url, data=payload, verify=self.verify)
-        return add_address_objects.json()["result"]
+            "params": [{"data": {
+                "allow-routing": allow_routing,
+                "associated-interface": associated_interface,
+                "name": name,
+                "subnet": subnet,
+                "type": object_type},
+                "url": f"pm/config/adom/{self.adom}/obj/firewall/address"}],
+            "session": self.sessionid}
+
+        add_address_object = session.post(url=self.base_url, json=payload, verify=self.verify)
+        return add_address_object.json()["result"]
+
+    def add_dynamic_object(self, name, device, subnet=list, comment=None):
+        """
+        Add per device mapping in address object.
+        :param name: name of the address object.
+        :param device: name of the device which is to be mapped in this object
+        :param subnet: subnet for device that is to be mapped in this object
+        :param comment: comment
+        :return: returns response of the request from FortiManager.
+        """
+        session = self.login()
+        add_obj = self.add_firewall_address_object(name, subnet=["0.0.0.0", "255.255.255.255"])
+        payload = {
+            "method": "add",
+            "params": [{"url": f"pm/config/adom/root/obj/firewall/address/{name}/dynamic_mapping",
+                        "data": [{"_scope": [{"name": f"{device}", "vdom": "root"}],
+                                  "subnet": subnet,
+                                  "comment": f"{comment}",
+                                  }]}],
+            "session": self.sessionid}
+        add_dynamic_obj = session.post(url=self.base_url, json=payload, verify=self.verify)
+        return [add_obj, add_dynamic_obj.json()["result"]]
+
+    def update_dynamic_object(self, name, device, subnet: list, do="add", comment=None):
+        """
+        Update the per mapping settings of the address object.
+        :param name: name of the object that needs to be updated.
+        :param device: name of the device that needs to be added/updated
+        :param subnet: updated subnet of the device that needs to be mapped
+        :param do: if parameter do is set to "add" it will update it. If it is set to "remove" it will be deleted.
+        :param comment: add comment if you want.
+        :return: return result of the request from FortiManager.
+        """
+        session = self.login()
+        payload = {
+            "params": [{"url": f"pm/config/adom/root/obj/firewall/address/{name}/dynamic_mapping",
+                        "data": [{"_scope": [{"name": f"{device}", "vdom": "root"}],
+                                  "subnet": subnet,
+                                  "comment": f"{comment}",
+                                  }]}],
+            "session": self.sessionid}
+        if do == "add":
+            payload.update(method="update")
+        elif do == "remove":
+            payload.update(method="delete")
+        update_dynamic_obj = session.post(url=self.base_url, json=payload, verify=self.verify)
+        return update_dynamic_obj.json()["result"]
 
     def update_firewall_address_object(self, name, **data):
         """
         Get the name of the address object and update it with your data
-
         :param name: Enter the name of the object that needs to be updated
         :param data: You can get the **kwargs parameters with "show_params_for_object_update()" method
         :return: Response of status code with data in JSON Format
@@ -232,7 +392,6 @@ class FortiManager:
     def delete_firewall_address_object(self, object_name):
         """
         Delete the address object if no longer needed using object name
-
         :param object_name: Enter the Object name you want to delete
         :return: Response of status code with data in JSON Format
         """
@@ -247,8 +406,7 @@ class FortiManager:
                 ],
                 "session": self.sessionid
             }
-        payload = repr(payload)
-        delete_address_object = session.post(url=self.base_url, data=payload, verify=self.verify)
+        delete_address_object = session.post(url=self.base_url, json=payload, verify=self.verify)
         return delete_address_object.json()["result"]
 
     # Firewall Address Groups Methods
@@ -272,8 +430,7 @@ class FortiManager:
                 ],
                 "session": self.sessionid
             }
-        payload = repr(payload)
-        get_address_group = session.post(url=self.base_url, data=payload, verify=self.verify)
+        get_address_group = session.post(url=self.base_url, json=payload, verify=self.verify)
         return get_address_group.json()["result"]
 
     def add_address_group(self, name, members=list):
@@ -298,8 +455,7 @@ class FortiManager:
                 ],
                 "session": self.sessionid
             }
-        payload = repr(payload)
-        add_address_group = session.post(url=self.base_url, data=payload, verify=self.verify)
+        add_address_group = session.post(url=self.base_url, json=payload, verify=self.verify)
         return add_address_group.json()["result"]
 
     def update_address_group(self, name, object_name, do="add"):
@@ -333,8 +489,7 @@ class FortiManager:
                 ],
                 "session": self.sessionid
             }
-        payload = repr(payload)
-        update_address_group = session.post(url=self.base_url, data=payload, verify=False)
+        update_address_group = session.post(url=self.base_url, json=payload, verify=False)
         return update_address_group.json()["result"]
 
     def delete_address_group(self, name):
@@ -356,15 +511,36 @@ class FortiManager:
                 ],
                 "session": self.sessionid
             }
-        payload = repr(payload)
-        delete_address_group = session.post(url=self.base_url, data=payload, verify=False)
+        delete_address_group = session.post(url=self.base_url, json=payload, verify=False)
         return delete_address_group.json()["result"]
 
+   # Firewall Virtual IP objects
+    def get_firewall_vip_objects(self, name=False):
+        """
+        Get all the vip objects data stored in FortiManager
+        :return: Response of status code with data in JSON Format
+        """
+        url = f"pm/config/adom/{self.adom}/obj/firewall/vip"
+        if name:
+            url = f"pm/config/adom/{self.adom}/obj/firewall/vip/{name}"
+        session = self.login()
+        payload = \
+            {
+                "method": "get",
+                "params": [
+                    {
+                        "url": url
+                    }
+                ],
+                "session": self.sessionid
+            }
+        get_vip_objects = session.post(url=self.base_url, json=payload, verify=self.verify)
+        return get_vip_objects.json()["result"]
+
     # Firewall Policies Methods
     def get_firewall_policies(self, policy_package_name="default", policyid=False):
         """
         Get the firewall policies present in the policy package
-
         :param policy_package_name: Enter the policy package name
         :param policyid: Can filter and get the policy you want using policyID
         :return: Response of status code with data in JSON Format
@@ -382,8 +558,7 @@ class FortiManager:
             ],
             "session": self.sessionid
         }
-        payload = repr(payload)
-        get_firewall_policies = session.post(url=self.base_url, data=payload, verify=self.verify)
+        get_firewall_policies = session.post(url=self.base_url, json=payload, verify=self.verify)
         return get_firewall_policies.json()["result"]
 
     def add_firewall_policy(self, policy_package_name="default", name=str, source_interface=str,
@@ -391,7 +566,6 @@ class FortiManager:
                             service=str, schedule="always", action=1, logtraffic=int, ):
         """
         Create your own policy in FortiManager using the instance parameters.
-
         :param policy_package_name: Enter the name of the policy package                eg. "default"
         :param name: Enter the policy name in a string format                           eg. "Test Policy"
         :param source_interface: Enter the source interface in a string format          eg. "port1"
@@ -428,14 +602,12 @@ class FortiManager:
             ],
             "session": self.sessionid
         }
-        payload = repr(payload)
-        add_policy = session.post(url=self.base_url, data=payload, verify=self.verify)
+        add_policy = session.post(url=self.base_url, json=payload, verify=self.verify)
         return add_policy.json()["result"]
 
     def update_firewall_policy(self, policy_package_name, policyid, **data):
         """
         Update your policy with your specific needs
-
         :param policy_package_name: Enter the policy package name in which you policy belongs
         :param policyid: Enter the Policy ID you want to edit
         :param data: You can get the **kwargs parameters with "show_params_for_policy_update()" method
@@ -454,14 +626,12 @@ class FortiManager:
                 ],
                 "session": self.sessionid
             }
-        payload = repr(payload)
-        update_policy = session.post(url=self.base_url, data=payload, verify=self.verify)
+        update_policy = session.post(url=self.base_url, json=payload, verify=self.verify)
         return update_policy.json()["result"]
 
     def delete_firewall_policy(self, policy_package_name, policyid):
         """
         Delete the policy if not is use with the policyID
-
         :param policy_package_name: Enter the policy package name in which you policy belongs
         :param policyid: Enter the policy ID of the policy you want to delete
         :return: Response of status code with data in JSON Format
@@ -477,14 +647,12 @@ class FortiManager:
                 ],
                 "session": self.sessionid
             }
-        payload = repr(payload)
-        delete_policy = session.post(url=self.base_url, data=payload, verify=self.verify)
+        delete_policy = session.post(url=self.base_url, json=payload, verify=self.verify)
         return delete_policy.json()["result"]
 
     def move_firewall_policy(self, policy_package_name, move_policyid=int, option="before", policyid=int):
         """
         Move the policy as per your needs
-
         :param policy_package_name: Enter the policy package name in which you policy belongs
         :param move_policyid: Enter the policy ID of the policy you want to move
         :param option: Specify if you want to move above("before") the target policy or below("after") {default: before}
@@ -504,14 +672,12 @@ class FortiManager:
                 ],
                 "session": self.sessionid
             }
-        payload = repr(payload)
-        move_policy = session.post(url=self.base_url, data=payload, verify=self.verify)
+        move_policy = session.post(url=self.base_url, json=payload, verify=self.verify)
         return move_policy.json()["result"]
 
     def install_policy_package(self, package_name):
         """
-        Install the policy package on your Fortigate Firewalls
-
+        Install the policy package on your Forti-gate Firewalls
         :param package_name: Enter the package name you wish to install
         :return: Response of status code with data in JSON Format
         """
@@ -530,8 +696,7 @@ class FortiManager:
                 ],
                 "session": self.sessionid
             }
-        payload = repr(payload)
-        install_package = session.post(url=self.base_url, data=payload, verify=self.verify)
+        install_package = session.post(url=self.base_url, json=payload, verify=self.verify)
         return install_package.json()["result"]
 
     @staticmethod
@@ -606,3 +771,18 @@ class FortiManager:
         comment(str)                    : Comments
         """
         return docs
+
+    def custom_api(self, payload):
+        """
+        Execute an API call manually by defining the payload
+        :param payload: specify the valid payload in a dict.
+        :return: returns response of the API call from FortiManager
+        """
+        session = self.login()
+        payload = payload
+        payload.update({"session": self.sessionid})
+        custom_api = session.post(url=self.base_url, json=payload, verify=self.verify)
+        return custom_api.json()
+
+    def set_adom(self, adom=None):
+        self.adom = adom
